/*
    Project: TCGINA
    Version: V1.21
    File   : kfi.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "kfipriv.h"

int KFI_InitEncryptionData(const KFI_DATA *pd, KFI_ENCRYPTION_DATA *ped,
  const void *pPassword, size_t cbPasswordSize)
{
    unsigned long uOutLen;
    int           iError;

    ped->fValid = FALSE;

    if (cbPasswordSize) {
        if (register_hash(&KFI_HASH_DESC) == -1) {
            return CRYPT_INVALID_HASH;
        }

        uOutLen = KFI_TOTAL_KEY_SIZE;
        if ((iError = pkcs_5_alg2(pPassword, (unsigned)cbPasswordSize,
          pd->rgbSalt, sizeof(pd->rgbSalt), KFI_ITERATION_COUNT,
          find_hash(KFI_HASH_NAME), (unsigned char*) ped, &uOutLen)) !=
          CRYPT_OK) {
             return iError;
        }
        if (uOutLen != KFI_TOTAL_KEY_SIZE) {
            return CRYPT_ERROR;
        }

        ped->fValid = TRUE;

        return CRYPT_OK;
    }

    return CRYPT_OK;
}

#ifdef KFI_DEBUG
extern const DWORD rgdwCRC32Table[256];
BOOL KFI_TestFunc(void)
{
    if (rijndael_test() != CRYPT_OK) {
        return FALSE;
    }
    if (twofish_test() != CRYPT_OK) {
        return FALSE;
    }

    if (calc_crc32("123456789", 9) != 0xcbf43926) {
        return FALSE;
    }
    if (calc_crc32(rgdwCRC32Table, sizeof(rgdwCRC32Table)) != 0x6fcf9e13) {
        return FALSE;
    }

#ifndef KFI_DECRYPTION_ONLY
    if (sha256_test() != CRYPT_OK) {
        return FALSE;
    }
#endif
    if (sha512_test() != CRYPT_OK) {
        return FALSE;
    }

    return TRUE;
}
#endif
