/*
    Project: TCGINA
    Version: V1.21
    File   : tcggina.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcgpriv.h"
#include <tchar.h>
#include <sddl.h>

/*-----------------------------------------------------------------------------
// GINA declarations                                                         */

typedef BOOL (WINAPI *PWLX_NEGOTIATE)(DWORD, DWORD *);
typedef BOOL (WINAPI *PWLX_INITIALIZE)(LPWSTR, HANDLE, PVOID, PVOID, PVOID *);
typedef VOID (WINAPI *PWLX_DISPLAY_SAS_NOTICE)(PVOID);
typedef int  (WINAPI *PWLX_LOGGED_OUT_SAS)(PVOID, DWORD, PLUID, PSID, PDWORD,
  PHANDLE, PWLX_MPR_NOTIFY_INFO, PVOID *);
typedef BOOL (WINAPI *PWLX_ACTIVATE_USERSHELL)(PVOID, PWSTR, PWSTR, PVOID);
typedef int  (WINAPI *PWLX_LOGGED_ON_SAS)(PVOID, DWORD, PVOID);
typedef VOID (WINAPI *PWLX_DISPLAY_LOCKED_NOTICE)(PVOID);
typedef int  (WINAPI *PWLX_WKSTA_LOCKED_SAS)(PVOID, DWORD);
typedef BOOL (WINAPI *PWLX_IS_LOCK_OK)(PVOID);
typedef BOOL (WINAPI *PWLX_IS_LOGOFF_OK)(PVOID);
typedef VOID (WINAPI *PWLX_LOGOFF)(PVOID);
typedef VOID (WINAPI *PWLX_SHUTDOWN)(PVOID, DWORD);

/* New for 1.1 */
typedef BOOL (WINAPI *PWLX_SCREEN_SAVER_NOTIFY)(PVOID, BOOL *);
typedef BOOL (WINAPI *PWLX_START_APPLICATION)(PVOID, PWSTR, PVOID, PWSTR);

/* New for 1.3 */
typedef BOOL (WINAPI *PWLX_NETWORK_PROVIDER_LOAD)(PVOID, PWLX_MPR_NOTIFY_INFO);
typedef BOOL (WINAPI *PWLX_DISPLAY_STATUS_MESSAGE)(PVOID, HDESK, DWORD, PWSTR,
  PWSTR);
typedef BOOL (WINAPI *PWLX_GET_STATUS_MESSAGE)(PVOID, DWORD *, PWSTR, DWORD);
typedef BOOL (WINAPI *PWLX_REMOVE_STATUS_MESSAGE)(PVOID);

/* New for 1.4 */
typedef BOOL (WINAPI *PWLX_GET_CONSOLE_SWITCH_CREDENTIALS)(PVOID, PVOID);
typedef VOID (WINAPI *PWLX_RECONNECT_NOTIFY)(PVOID);
typedef VOID (WINAPI *PWLX_DISCONNECT_NOTIFY)(PVOID);

/*-----------------------------------------------------------------------------
// Global variables                                                          */

HINSTANCE g_hinstDLL;
static DWORD g_dwNegotiatedWinlogonVersion;
static TCG_ENCRYPTED_PROFILE_INFO g_epiDefault;

static PWLX_NEGOTIATE pfWlxNegotiate;
static PWLX_INITIALIZE pfWlxInitialize;
static PWLX_DISPLAY_SAS_NOTICE pfWlxDisplaySASNotice;
static PWLX_LOGGED_OUT_SAS pfWlxLoggedOutSAS;
static PWLX_ACTIVATE_USERSHELL pfWlxActivateUserShell;
static PWLX_LOGGED_ON_SAS pfWlxLoggedOnSAS;
static PWLX_DISPLAY_LOCKED_NOTICE pfWlxDisplayLockedNotice;
static PWLX_WKSTA_LOCKED_SAS pfWlxWkstaLockedSAS;
static PWLX_IS_LOCK_OK pfWlxIsLockOk;
static PWLX_IS_LOGOFF_OK pfWlxIsLogoffOk;
static PWLX_LOGOFF pfWlxLogoff;
static PWLX_SHUTDOWN pfWlxShutdown;

/* New for 1.1 */
static PWLX_START_APPLICATION pfWlxStartApplication;
static PWLX_SCREEN_SAVER_NOTIFY pfWlxScreenSaverNotify;

/* New for 1.3 */
static PWLX_NETWORK_PROVIDER_LOAD pfWlxNetworkProviderLoad;
static PWLX_DISPLAY_STATUS_MESSAGE pfWlxDisplayStatusMessage;
static PWLX_GET_STATUS_MESSAGE pfWlxGetStatusMessage;
static PWLX_REMOVE_STATUS_MESSAGE pfWlxRemoveStatusMessage;

/* New for 1.4 */
static PWLX_GET_CONSOLE_SWITCH_CREDENTIALS pfWlxGetConsoleSwitchCredentials;
static PWLX_RECONNECT_NOTIFY pfWlxReconnectNotify;
static PWLX_DISCONNECT_NOTIFY pfWlxDisconnectNotify;

/*-----------------------------------------------------------------------------
// DllMain                                                                   */

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD dwReason, LPVOID lpReserved)
{
    UNREFERENCED_PARAMETER(lpReserved);

    if (dwReason == DLL_PROCESS_ATTACH) {
        TC_DbgSetHInstance(hinstDLL);
        g_hinstDLL = hinstDLL;
        g_dwNegotiatedWinlogonVersion = 0;
    }
    if (dwReason == DLL_PROCESS_DETACH) {
        TC_DbgPutsA("DLL_PROCESS_DETACH\n");
    }

    return TRUE;
}

#define LOAD_FUNCTION(h, x) \
  ((*(FARPROC*)&pf##x = GetProcAddress(h, #x)) != NULL)

static BOOL WINAPI TCG_LoadGinaDLL(LPCTSTR pszFileName, HMODULE *phDll)
{
    HMODULE hDll;

    TC_Assert(pszFileName);

    if (*pszFileName == TEXT('\0')) return FALSE;
    if ((hDll = LoadLibrary(pszFileName)) == NULL) {
        TC_DbgPrintf(TEXT("Cannot load GINA DLL \"%s\"\n"), pszFileName);
        return FALSE;
    }

    if (!LOAD_FUNCTION(hDll, WlxNegotiate)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxInitialize)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxDisplaySASNotice)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxLoggedOutSAS)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxActivateUserShell)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxLoggedOnSAS)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxDisplayLockedNotice)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxIsLockOk)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxWkstaLockedSAS)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxIsLogoffOk)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxLogoff)) goto error;
    if (!LOAD_FUNCTION(hDll, WlxShutdown)) goto error;

    LOAD_FUNCTION(hDll, WlxStartApplication);
    LOAD_FUNCTION(hDll, WlxScreenSaverNotify);

    LOAD_FUNCTION(hDll, WlxNetworkProviderLoad);
    LOAD_FUNCTION(hDll, WlxDisplayStatusMessage);
    LOAD_FUNCTION(hDll, WlxGetStatusMessage);
    LOAD_FUNCTION(hDll, WlxRemoveStatusMessage);

    LOAD_FUNCTION(hDll, WlxGetConsoleSwitchCredentials);
    LOAD_FUNCTION(hDll, WlxReconnectNotify);
    LOAD_FUNCTION(hDll, WlxDisconnectNotify);

    if (phDll != NULL) {
        *phDll = hDll;
    }

    return TRUE;

error:
    FreeLibrary(hDll);
    TC_DbgPrintf(TEXT("Cannot load GINA DLL \"%s\"\n"), pszFileName);
    return FALSE;
}

/*-----------------------------------------------------------------------------
// WinLogon functions                                                        */

static PTCG_CONTEXT g_pFirstCtx = NULL;
static WLX_DISPATCH_VERSION_1_4 g_RedirectedWinlogonFunctions;

VOID WINAPI WlxSetContextPointer(HANDLE hWlx, PVOID pWlxContext)
{
    TC_DbgPutsA("WlxSetContextPointer\n");

    if (hWlx == g_pFirstCtx->hWlx) {
        g_pFirstCtx->pMsginaContext = pWlxContext;
    } else {
        g_pFirstCtx->pWlxFuncs->WlxSetContextPointer(hWlx, pWlxContext);
    }
}

BOOL WINAPI WlxSetOption(HANDLE hWlx, DWORD dwOption, ULONG_PTR ulpValue,
  ULONG_PTR* pulpOldValue)
{
    TC_DbgPutsA("WlxSetOption\n");

    if (hWlx == g_pFirstCtx->hWlx) {
        if (dwOption == WLX_OPTION_CONTEXT_POINTER) {
            if (pulpOldValue != NULL) {
                *pulpOldValue = (ULONG_PTR) g_pFirstCtx->pMsginaContext;
            }
            g_pFirstCtx->pMsginaContext = (PVOID) ulpValue;
            return TRUE;
        }
    }

    return g_pFirstCtx->pWlxFuncs->WlxSetOption(hWlx, dwOption, ulpValue,
      pulpOldValue);
}

BOOL WINAPI WlxGetOption(HANDLE hWlx, DWORD dwOption, ULONG_PTR* pulpValue)
{
    TC_DbgPutsA("WlxGetOption\n");

    if (hWlx == g_pFirstCtx->hWlx) {
        if (dwOption == WLX_OPTION_CONTEXT_POINTER) {
            if (pulpValue == NULL) return FALSE;
            *pulpValue = (ULONG_PTR) g_pFirstCtx->pMsginaContext;
            return TRUE;
        }
    }

    return g_pFirstCtx->pWlxFuncs->WlxGetOption(hWlx, dwOption, pulpValue);
}

/*-----------------------------------------------------------------------------
// GINA functions                                                            */

BOOL WINAPI WlxNegotiate(DWORD dwWinlogonVersion, PDWORD pdwDllVersion)
{
    TCHAR szFileName[MAX_PATH];

    TC_Assert(pdwDllVersion != NULL);

    TC_DbgPutsA("WlxNegotiate\n");

    g_dwNegotiatedWinlogonVersion = 0;

    *szFileName = TEXT('\0');
    TC_GetRegistryString(HKEY_LOCAL_MACHINE, TCG_REGKEY_NAME,
      TEXT("GinaDLL"), szFileName, MAX_PATH);
    if (*szFileName != TEXT('\0')) {
        if (!TCG_LoadGinaDLL(szFileName, NULL)) {
            return FALSE;
        }
    } else {
        HINSTANCE hTargetDll;
        if (!TCG_LoadGinaDLL(TEXT("MSGINA.DLL"), &hTargetDll)) {
            return FALSE;
        }
        if (TC_GetRegistryInteger(HKEY_LOCAL_MACHINE, TCG_REGKEY_NAME,
          TEXT("HideGinaDLL"), 0)) {
            InstallHook(hTargetDll, g_hinstDLL);
        }
    }

    if (dwWinlogonVersion > WLX_VERSION_1_4) {
        dwWinlogonVersion = WLX_VERSION_1_4;
    }
    if (!pfWlxNegotiate(dwWinlogonVersion, pdwDllVersion)) {
        return FALSE;
    }

    g_dwNegotiatedWinlogonVersion = *pdwDllVersion;

    return TRUE;
}

BOOL WINAPI WlxInitialize(LPWSTR lpWinsta, HANDLE hWlx, PVOID pvReserved,
  PVOID pWinlogonFunctions, PVOID *pWlxContext)
{
    static BOOL   fRunOnceAlreadyExecuted = FALSE;
    PTCG_CONTEXT  pCtx;
    OSVERSIONINFO ovi;

    TC_Assert(pfWlxInitialize != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxInitialize\n");

    pCtx = (PTCG_CONTEXT) TC_Calloc(sizeof(TCG_CONTEXT));
    if (!pCtx) {
        return FALSE;
    }

    pCtx->lpWinsta = lpWinsta;
    pCtx->hWlx = hWlx;
    pCtx->pWlxFuncs = (PWLX_DISPATCH_VERSION_1_4) pWinlogonFunctions;
    pCtx->fRemoteSession = GetSystemMetrics(SM_REMOTESESSION) != 0;

    ovi.dwOSVersionInfoSize = sizeof(ovi);
    if (GetVersionEx(&ovi)) {
        pCtx->dwOSVersion =
          (ovi.dwMajorVersion << 16) | (ovi.dwMinorVersion & 0xffff);
    }

#define REDIRECT(x) g_RedirectedWinlogonFunctions.##x = x
    if (g_pFirstCtx == NULL) {
        g_pFirstCtx = pCtx;
        memcpy(&g_RedirectedWinlogonFunctions, pWinlogonFunctions,
          sizeof(g_RedirectedWinlogonFunctions));
        REDIRECT(WlxSetContextPointer);
        REDIRECT(WlxSetOption);
        REDIRECT(WlxGetOption);
        pWinlogonFunctions = &g_RedirectedWinlogonFunctions;
    }
#undef REDIRECT

    if (!pfWlxInitialize(lpWinsta, hWlx, pvReserved, pWinlogonFunctions,
      &pCtx->pMsginaContext)) {
        TC_DbgPutsA("MSGINA.WlxInitialize failed!\n");
        LocalFree(pCtx);
        g_pFirstCtx = NULL;

        return FALSE;
    }
    *pWlxContext = pCtx;

    if (!fRunOnceAlreadyExecuted) {
        fRunOnceAlreadyExecuted = TRUE;
        memset(&g_epiDefault, '\0', sizeof(g_epiDefault));
        pCtx->pProfileBegin = &g_epiDefault;
        pCtx->pProfileEnd = pCtx->pProfileBegin + 1;
        pCtx->dwMSOptions = TC_GetRegistryInteger(HKEY_LOCAL_MACHINE,
          TCG_REGKEY_NAME, TEXT("MSOptions"), 0);
#ifdef SUPPORT_SHARED_MEMORY
        pCtx->dwMSCount = TC_GetRegistryInteger(HKEY_LOCAL_MACHINE,
          TCG_REGKEY_NAME, TEXT("MSCount"), 0);
        if (pCtx->dwMSCount) {
            pCtx->pShareMemory = TCG_InitSharedMemory(&pCtx->dwMSCount);
            if (pCtx->pShareMemory) {
                pCtx->pProfileBegin = pCtx->pShareMemory->rgProfileInfo;
                pCtx->pProfileEnd = pCtx->pProfileBegin + pCtx->dwMSCount;
            }
        }
#endif

        if (!pCtx->fRemoteSession) {
            TC_HANDLE hTC;
            TCHAR     szDrives[26+1];
            pCtx->dwProtectedDrives = 0;
            if (TC_GetRegistryString(HKEY_LOCAL_MACHINE,
              TCG_REGKEY_NAME, TEXT("ProtectedDrives"), szDrives, countof(szDrives))) {
                LPTSTR psz = szDrives;
                _tcsupr(szDrives);
                while (*psz) {
                    pCtx->dwProtectedDrives |= 1 << (*psz - TEXT('A'));
                    psz++;
                }
            }
            hTC = TC_Open(0);
            if (hTC) {
                if (!TC_GetMountedDrives(hTC, &pCtx->dwPreMountedDrives, TCMD_ALL)) {
                    pCtx->dwPreMountedDrives = 0;
                }
                TC_Close(hTC);
            }
            pCtx->dwProtectedDrives |= pCtx->dwPreMountedDrives;
            /* add auto mounted drives later */
#ifdef SUPPORT_SHARED_MEMORY
            if (TCG_LockSharedMemory()) {
                pCtx->pShareMemory->dwPreMountedDrives =
                  pCtx->dwPreMountedDrives;
                pCtx->pShareMemory->dwProtectedDrives =
                  pCtx->dwProtectedDrives;
                TCG_UnlockSharedMemory();
            }
#endif
        } else {
#ifdef SUPPORT_AUTOMOUNT
            pCtx->fAutoMountDone = TRUE;
#endif
#ifdef SUPPORT_SHARED_MEMORY
            if (TCG_LockSharedMemory()) {
                pCtx->dwPreMountedDrives =
                  pCtx->pShareMemory->dwPreMountedDrives;
                pCtx->dwProtectedDrives =
                  pCtx->pShareMemory->dwProtectedDrives;
                TCG_UnlockSharedMemory();
            }
#endif
        }
    }

    return TRUE;
}

VOID WINAPI WlxDisplaySASNotice(PVOID pWlxContext)
{
    TC_Assert(pfWlxDisplaySASNotice != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxDisplaySASNotice\n");

    pfWlxDisplaySASNotice(GetMsginaContext(pWlxContext));
}

int WINAPI WlxLoggedOutSAS(PVOID pWlxContext, DWORD dwSasType,
  PLUID pAuthenticationId, PSID pLogonSid, PDWORD pdwOptions,
  PHANDLE phToken, PWLX_MPR_NOTIFY_INFO pMprNotifyInfo, PVOID *ppProfile)
{
    int          iResult;
    PTCG_CONTEXT pCtx;

    TC_Assert(pfWlxLoggedOutSAS != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxLoggedOutSAS\n");

    pCtx = (PTCG_CONTEXT) pWlxContext;
#ifdef SUPPORT_AUTOMOUNT
    if (!pCtx->fAutoMountDone) {
        DWORD dwProtectedDrives;
        TC_HANDLE hTC;
        if (!TCG_Automount(pCtx)) {
            return WLX_SAS_ACTION_NONE;
        }
        pCtx->fAutoMountDone = TRUE;
        hTC = TC_Open(0);
        if (hTC) {
            if (!TC_GetMountedDrives(hTC, &dwProtectedDrives, TCMD_ALL)) {
                dwProtectedDrives = 0;
            }
            TC_Close(hTC);
            pCtx->dwProtectedDrives |= dwProtectedDrives;
#ifdef SUPPORT_SHARED_MEMORY
            if (TCG_LockSharedMemory()) {
                pCtx->pShareMemory->dwProtectedDrives =
                  pCtx->dwProtectedDrives;
                TCG_UnlockSharedMemory();
            }
#endif
        }
    }
#endif

    TCG_LoggedOff(pCtx);

#ifdef TC_DEBUG
    {
        LPTSTR pszStringSid;
        if (ConvertSidToStringSid(pLogonSid, &pszStringSid)) {
            TC_DbgPrintf(TEXT("LogonSid: %s\n"), pszStringSid);
            LocalFree(pszStringSid);
        }
    }
#endif

    iResult = pfWlxLoggedOutSAS(GetMsginaContext(pWlxContext), dwSasType,
      pAuthenticationId, pLogonSid, pdwOptions, phToken, pMprNotifyInfo,
      ppProfile);

    if (iResult != WLX_SAS_ACTION_LOGON) return iResult;

#ifdef TC_DEBUG
    {
        LPTSTR pszStringSid;
        if (ConvertSidToStringSid(pLogonSid, &pszStringSid)) {
            TC_DbgPrintf(TEXT("LogonSid: %s\n"), pszStringSid);
            LocalFree(pszStringSid);
        }
    }
    TC_DbgPrintfA("Authentication ID: 0x%016I64x\n",
      *(unsigned __int64*)pAuthenticationId);
    if (pdwOptions) {
        TC_DbgPrintfA("dwOptions: 0x%08x\n", (unsigned)*pdwOptions);
    }
    if (pMprNotifyInfo) {
        if (pMprNotifyInfo->pszUserName) {
            TC_DbgPrintfW(L"User name: %s\n", pMprNotifyInfo->pszUserName);
        }
        if (pMprNotifyInfo->pszDomain) {
            TC_DbgPrintfW(L"Domain: %s\n", pMprNotifyInfo->pszDomain);
        }
    }
    if (ppProfile) {
        WLX_PROFILE_V1_0 **ppProfile1 = (WLX_PROFILE_V1_0**)ppProfile;
        if (*ppProfile) {
            switch ((*ppProfile1)->dwType) {
            case WLX_PROFILE_TYPE_V1_0: {
                if ((*ppProfile1)->pszProfile) {
                    TC_DbgPrintfW(L"Profile: %s\n", (*ppProfile1)->pszProfile);
                }
                break; }
            case WLX_PROFILE_TYPE_V2_0: {
                WLX_PROFILE_V2_0 **ppProfile2 = (WLX_PROFILE_V2_0**)ppProfile;
                if ((*ppProfile2)->pszProfile) {
                    TC_DbgPrintfW(L"Profile: %s\n", (*ppProfile2)->pszProfile);
                }
                if ((*ppProfile2)->pszPolicy) {
                    TC_DbgPrintfW(L"Policy: %s\n", (*ppProfile2)->pszPolicy);
                }
                if ((*ppProfile2)->pszNetworkDefaultUserProfile) {
                    TC_DbgPrintfW(L"Network default user profile: %s\n",
                      (*ppProfile2)->pszNetworkDefaultUserProfile);
                }
                if ((*ppProfile2)->pszServerName) {
                    TC_DbgPrintfW(L"Server name: %s\n",
                      (*ppProfile2)->pszServerName);
                }
                if ((*ppProfile2)->pszEnvironment) {
                    PWSTR pwStr = (*ppProfile2)->pszEnvironment;
                    for(; *pwStr; pwStr += wcslen(pwStr) + 1) {
                        TC_DbgPrintfW(L"Environment: %s\n", pwStr);
                    }
                }
                break; }
            }
        }
    }
#endif

    if (pMprNotifyInfo == NULL || pMprNotifyInfo->pszUserName == NULL) {
        TC_DbgPutsA("Unknown user name!\n");
        return WLX_SAS_ACTION_LOGON;
    }

    if (!TCG_Logon(pCtx, pMprNotifyInfo->pszUserName,
      pMprNotifyInfo->pszPassword, pMprNotifyInfo->pszDomain)) {
        /* simulate a user logoff */
        iResult = WLX_SAS_ACTION_NONE;
        if (*phToken) {
            CloseHandle(*phToken);
            *phToken = 0;
        }
        if (ppProfile) {
            WLX_PROFILE_V1_0 **ppProfile1 = (WLX_PROFILE_V1_0**)ppProfile;
            if (*ppProfile) {
                switch ((*ppProfile1)->dwType) {
                case WLX_PROFILE_TYPE_V1_0: {
                    LocalFree((*ppProfile1)->pszProfile);
                    LocalFree(*ppProfile1);
                    *ppProfile1 = NULL;
                    break; }
                case WLX_PROFILE_TYPE_V2_0: {
                    WLX_PROFILE_V2_0 **ppProfile2 = (WLX_PROFILE_V2_0**)ppProfile;
                    LocalFree((*ppProfile2)->pszEnvironment);
                    LocalFree((*ppProfile2)->pszNetworkDefaultUserProfile);
                    LocalFree((*ppProfile2)->pszPolicy);
                    LocalFree((*ppProfile2)->pszProfile);
                    LocalFree((*ppProfile2)->pszServerName);
                    LocalFree(*ppProfile2);
                    *ppProfile2 = NULL;
                    break; }
                }
            }
        }
        if (pMprNotifyInfo) {
            LocalFree(pMprNotifyInfo->pszDomain);
            LocalFree(pMprNotifyInfo->pszOldPassword);
            LocalFree(pMprNotifyInfo->pszPassword);
            LocalFree(pMprNotifyInfo->pszUserName);
            TC_ZeroMemory(pMprNotifyInfo, sizeof(*pMprNotifyInfo));
        }
        if (WlxIsLogoffOk(pWlxContext)) {
            pfWlxLogoff(GetMsginaContext(pWlxContext));
        }
    }

    return iResult;
}

BOOL WINAPI WlxActivateUserShell(PVOID pWlxContext, PWSTR pszDesktopName,
  PWSTR pszMprLogonScript, PVOID pEnvironment)
{
    TC_Assert(pfWlxActivateUserShell != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxActivateUserShell\n");

    return pfWlxActivateUserShell(GetMsginaContext(pWlxContext),
      pszDesktopName, pszMprLogonScript, pEnvironment);
}

int WINAPI WlxLoggedOnSAS(PVOID pWlxContext, DWORD dwSasType, PVOID pReserved)
{
    TC_Assert(pfWlxLoggedOnSAS != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxLoggedOnSAS\n");

    return pfWlxLoggedOnSAS(GetMsginaContext(pWlxContext), dwSasType,
      pReserved);
}

VOID WINAPI WlxDisplayLockedNotice(PVOID pWlxContext)
{
    TC_Assert(pfWlxDisplayLockedNotice != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxDisplayLockedNotice\n");


    pfWlxDisplayLockedNotice(GetMsginaContext(pWlxContext));
}

BOOL WINAPI WlxIsLockOk(PVOID pWlxContext)
{
    TC_Assert(pfWlxIsLockOk != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxIsLockOk\n");

    return pfWlxIsLockOk(GetMsginaContext(pWlxContext));
}

int WINAPI WlxWkstaLockedSAS(PVOID pWlxContext, DWORD dwSasType)
{
    TC_Assert(pfWlxWkstaLockedSAS != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxWkstaLockedSAS\n");

    return pfWlxWkstaLockedSAS(GetMsginaContext(pWlxContext), dwSasType);
}

BOOL WINAPI WlxIsLogoffOk(PVOID pWlxContext)
{
    TC_Assert(pfWlxIsLogoffOk != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxIsLogoffOk\n");

    return pfWlxIsLogoffOk(GetMsginaContext(pWlxContext));
}

VOID WINAPI WlxLogoff(PVOID pWlxContext)
{
    TC_Assert(pfWlxLogoff != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxLogoff\n");

    pfWlxLogoff(GetMsginaContext(pWlxContext));

    TCG_Logoff((PTCG_CONTEXT) pWlxContext);
}

static VOID WINAPI TCG_DismountAllNonPreMountedVolumes(PTCG_CONTEXT pCtx)
{
    TC_HANDLE    hTC;

    TC_Assert(pCtx != NULL);

    hTC = TC_Open(0);
    if (hTC) {
        DWORD dwMountedDrives;
        if (!TC_GetMountedDrives(hTC, &dwMountedDrives,
          TCMD_DISMOUNTABLE|TCMD_PERSISTENT)) {
            dwMountedDrives = 0x3ffffff;
        }
        dwMountedDrives &= ~pCtx->dwPreMountedDrives;
        if (dwMountedDrives) {
            INT   i;
            TCHAR szDrive[3] = TEXT("A:");
            if (pCtx->fFlushRegistry) {
                LONG  lResult;
                TC_DbgPutsA("Flush HKEY_USERS\n");
                lResult = RegFlushKey(HKEY_USERS);
                if (lResult != ERROR_SUCCESS) {
                    TC_DbgPrintfA("Unsuccessful: Win32 Error Code 0x%04x\n", lResult);
                }
            }
            for (i = 0; i < 26; i++) {
                if (dwMountedDrives & (1<<i)) {
                    szDrive[0] = TEXT('A') + (TCHAR)i;
                    TC_Dismount(hTC, szDrive, TC_DISMOUNT_FORCIBLY);
                }
            }
        }
        TC_Close(hTC);
    }
}

VOID WINAPI WlxShutdown(PVOID pWlxContext, DWORD ShutdownType)
{
    PTCG_CONTEXT pCtx;

    TC_Assert(pfWlxShutdown != NULL);
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxShutdown\n");

    pfWlxShutdown(GetMsginaContext(pWlxContext), ShutdownType);

    pCtx = (PTCG_CONTEXT) pWlxContext;

#ifdef SUPPORT_SHARED_MEMORY
    TC_Assert(!pCtx->fRemoteSession);
    if (!pCtx->fRemoteSession && pCtx->dwMSCount) {
        if (g_dwNegotiatedWinlogonVersion >= WLX_VERSION_1_3) {
            /* Defer dismount to WlxDisplayStatusMessage
             * because remote sessions are possibly still running
             */
            pCtx->fShuttingDown = TRUE;
            return;
        }
    }
#endif

    TCG_DismountAllNonPreMountedVolumes(pCtx);
}

/* New for 1.1 */

BOOL WINAPI WlxScreenSaverNotify(PVOID pWlxContext, BOOL *pSecure)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxScreenSaverNotify\n");

    if (pfWlxScreenSaverNotify) {
        return pfWlxScreenSaverNotify(GetMsginaContext(pWlxContext), pSecure);
    }

    if (pSecure) *pSecure = TRUE;

    return TRUE;
}

BOOL WINAPI WlxStartApplication(PVOID pWlxContext, PWSTR pszDesktopName,
  PVOID pEnvironment, PWSTR pszCmdLine)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxStartApplication\n");

    if (pfWlxStartApplication) {
        return pfWlxStartApplication(GetMsginaContext(pWlxContext),
          pszDesktopName, pEnvironment, pszCmdLine);
    }

    return FALSE;
}

/* New for 1.3 */

BOOL WINAPI WlxNetworkProviderLoad(PVOID pWlxContext,
  PWLX_MPR_NOTIFY_INFO pNprNotifyInfo)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxNetworkProviderLoad\n");

    if (pfWlxNetworkProviderLoad) {
        return pfWlxNetworkProviderLoad(GetMsginaContext(pWlxContext),
          pNprNotifyInfo);
    }

    return FALSE;
}

BOOL WINAPI WlxDisplayStatusMessage(PVOID pWlxContext, HDESK hDesktop,
  DWORD dwOptions, PWSTR pTitle, PWSTR pMessage)
{
    PTCG_CONTEXT pCtx;

    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxDisplayStatusMessage\n");

    pCtx = (PTCG_CONTEXT) pWlxContext;
    if (!pCtx->fRemoteSession && pCtx->fShuttingDown) {
        pCtx->fShuttingDown = FALSE;
        TCG_DismountAllNonPreMountedVolumes(pCtx);
    }

    if (pfWlxDisplayStatusMessage) {
        return pfWlxDisplayStatusMessage(GetMsginaContext(pWlxContext),
          hDesktop, dwOptions, pTitle, pMessage);
    }

    return FALSE;
}

BOOL WINAPI WlxGetStatusMessage(PVOID pWlxContext, DWORD *pdwOptions,
  PWSTR pMessage, DWORD dwBufferSize)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxGetStatusMessage\n");

    if (pfWlxGetStatusMessage) {
        return pfWlxGetStatusMessage(GetMsginaContext(pWlxContext),
          pdwOptions, pMessage, dwBufferSize);
    }

    return FALSE;
}

BOOL WINAPI WlxRemoveStatusMessage(PVOID pWlxContext)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxRemoveStatusMessage\n");

    if (pfWlxRemoveStatusMessage) {
        return pfWlxRemoveStatusMessage(GetMsginaContext(pWlxContext));
    }

    return FALSE;
}

/* New for 1.4 */

BOOL WINAPI WlxGetConsoleSwitchCredentials(PVOID pWlxContext,
  PVOID pCredInfo)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxGetConsoleSwitchCredentials\n");

    if (pfWlxGetConsoleSwitchCredentials) {
        return pfWlxGetConsoleSwitchCredentials(GetMsginaContext(pWlxContext),
          pCredInfo);
    }

    return FALSE;
}

VOID WINAPI WlxReconnectNotify(PVOID pWlxContext)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxReconnectNotify\n");

    if (pfWlxReconnectNotify) {
        pfWlxReconnectNotify(GetMsginaContext(pWlxContext));
    }
}

VOID WINAPI WlxDisconnectNotify(PVOID pWlxContext)
{
    TC_Assert(pWlxContext != NULL);

    TC_DbgPutsA("WlxDisconnectNotify\n");

    if (pfWlxDisconnectNotify) {
        pfWlxDisconnectNotify(GetMsginaContext(pWlxContext));
    }
}
